﻿using Mp4SubtitleParser;
using N_m3u8DL_RE.Common.Log;
using N_m3u8DL_RE.Common.Resource;
using System.Diagnostics;
using System.Text.RegularExpressions;
using N_m3u8DL_RE.Enum;

namespace N_m3u8DL_RE.Util;

internal static partial class MP4DecryptUtil
{
    private static readonly string ZeroKid = "00000000000000000000000000000000";
    public static async Task<bool> DecryptAsync(DecryptEngine decryptEngine, string bin, string[]? keys, string source, string dest, string? kid, string init = "", bool isMultiDRM=false)
    {
        if (keys == null || keys.Length == 0) return false;

        var keyPairs = keys.ToList();
        string? keyPair = null;
        string? trackId = null;
        string? tmpEncFile = null;
        string? tmpDecFile = null;
        string? workDir = null;

        if (isMultiDRM)
        {
            trackId = "1";
        }

        if (!string.IsNullOrEmpty(kid))
        {
            var test = keyPairs.Where(k => k.StartsWith(kid)).ToList();
            if (test.Count != 0) keyPair = test.First();
        }

        // Apple
        if (kid == ZeroKid)
        {
            keyPair = keyPairs.First();
            trackId = "1";
        }

        // user only input key, append kid
        if (keyPair == null && keyPairs.Count == 1 && !keyPairs.First().Contains(':'))
        {
            keyPairs = keyPairs.Select(x => $"{kid}:{x}").ToList();
            keyPair = keyPairs.First();
        }
            
        if (keyPair == null) return false;

        // shakaPackager/ffmpeg 无法单独解密init文件
        if (source.EndsWith("_init.mp4") && decryptEngine != DecryptEngine.MP4DECRYPT) return false;

        string cmd = "";

        var tmpFile = "";
        if (decryptEngine == DecryptEngine.SHAKA_PACKAGER)
        {
            var enc = source;
            // shakaPackager 手动构造文件
            if (init != "")
            {
                tmpFile = Path.ChangeExtension(source, ".itmp");
                MergeUtil.CombineMultipleFilesIntoSingleFile([init, source], tmpFile);
                enc = tmpFile;
            }

            cmd = $"--quiet --enable_raw_key_decryption input=\"{enc}\",stream=0,output=\"{dest}\" " +
                  $"--keys {(trackId != null ? $"label={trackId}:" : "")}key_id={(trackId != null ? ZeroKid : kid)}:key={keyPair.Split(':')[1]}";
        }
        else if (decryptEngine == DecryptEngine.MP4BOX)
        {
            // Logger.Info($"decryptEngine: {decryptEngine}"); // decryptEngine: MP4BOX
            // Logger.Info($"bin: {bin}"); // bin: /Applications/GPAC.app/Contents/MacOS/MP4Box
            // Logger.Info($"keys: {string.Join(", ", keys ?? new string[0])}"); // keys: ccbf5fb4c2965be7aa130ffb3ba9fd73:9cc0c92044cb1d69433f5f5839a159df, 9bf0e9cf0d7b55aeb4b289a63bab8610:90f52fd8ca48717b21d0c2fed7a12ae1, eb676abbcb345e96bbcf616630f1a3da:100b6c20940f779a4589152b57d2dacb, 0294b9599d755de2bbf0fdca3fa5eab7:3bda2f40344c7def614227b9c0f03e26, 639da80cf23b55f3b8cab3f64cfa5df6:229f5f29b643e203004b30c4eaf348f4
            // Logger.Info($"source: {source}"); // source: /Users/noham/Documents/clone/N_m3u8DL-RE/artifact-arm64/11331_2025-02-01_20-04-27.mp4
            // Logger.Info($"dest: {dest}"); // dest: /Users/noham/Documents/clone/N_m3u8DL-RE/artifact-arm64/11331_2025-02-01_20-04-27_dec.mp4
            // Logger.Info($"kid: {kid}"); // kid: eb676abbcb345e96bbcf616630f1a3da
            // Logger.Info($"init: {init}"); // init: 
            // Logger.Info($"isMultiDRM: {isMultiDRM}"); // isMultiDRM: False 
            // Logger.Info($"trackId: {trackId}"); // trackId: 
            // Logger.Info($"keyPair: {keyPair}"); // keyPair: eb676abbcb345e96bbcf616630f1a3da:100b6c20940f779a4589152b57d2dacb
            // Logger.Info($"tmpEncFile: {tmpEncFile}"); // tmpEncFile: 
            // Logger.Info($"tmpDecFile: {tmpDecFile}"); // tmpDecFile:
            // Logger.Info($"workDir: {workDir}"); // workDir:
            // Logger.Info($"tmpFile: {tmpFile}"); // tmpFile:

            var name = Path.GetFileNameWithoutExtension(source);
            // Logger.Info($"name: {name}");

            workDir = Path.GetDirectoryName(source)!;

            var enc = source;
            if (init != "")
            {
                var mergedinitenc = Path.ChangeExtension(source, ".itmp");
                MergeUtil.CombineMultipleFilesIntoSingleFile([init, source], mergedinitenc);
                enc = mergedinitenc;
            }
            // Logger.Info($"enc: {enc}");

            var xmlfile = Path.Combine(workDir, $"{name}_drm.xml.itmp");
            tmpFile = xmlfile;
            var xmlContent = $"""
            <?xml version="1.0" encoding="UTF-8"?>
            <GPACDRM type="CENC">
                <CrypTrack trackID="1">
                    <key KID="{keyPair.Split(':')[0]}" value="{keyPair.Split(':')[1]}"/>
                </CrypTrack>
            </GPACDRM>
            """;
            File.WriteAllText(xmlfile, xmlContent);
            
            // Logger.Info($"xmlfile: {xmlfile}");
            // Logger.Info($"Decrypting {Path.GetFileName(enc)} using MP4Box");
            cmd = $"-decrypt \"{xmlfile}\" \"{enc}\" -out \"{dest}\" -logs=all@error"; // f"MP4Box -decrypt {xmlfile} {file} -out {file} -logs=all@error"
            
            
        }
        else if (decryptEngine == DecryptEngine.MP4DECRYPT)
        {
            if (trackId == null)
            {
                cmd = string.Join(" ", keyPairs.Select(k => $"--key {k}"));
            }
            else
            {
                cmd = string.Join(" ", keyPairs.Select(k => $"--key {trackId}:{k.Split(':')[1]}"));
            }
            // 解决mp4decrypt中文问题 切换到源文件所在目录并改名再解密
            workDir = Path.GetDirectoryName(source)!;
            tmpEncFile = Path.Combine(workDir, $"{Guid.NewGuid()}{Path.GetExtension(source)}");
            tmpDecFile = Path.Combine(workDir, $"{Path.GetFileNameWithoutExtension(tmpEncFile)}_dec{Path.GetExtension(tmpEncFile)}");
            File.Move(source, tmpEncFile);
            if (init != "")
            {
                var infoFile = Path.GetDirectoryName(init) == workDir ? Path.GetFileName(init) : init;
                cmd += $" --fragments-info \"{infoFile}\" ";
            }
            cmd += $" \"{Path.GetFileName(tmpEncFile)}\" \"{Path.GetFileName(tmpDecFile)}\"";
        }
        else
        {
            var enc = source;
            // ffmpeg实时解密 手动构造文件
            if (init != "")
            {
                tmpFile = Path.ChangeExtension(source, ".itmp");
                MergeUtil.CombineMultipleFilesIntoSingleFile([init, source], tmpFile);
                enc = tmpFile;
            }
            
            cmd = $"-loglevel error -nostdin -decryption_key {keyPair.Split(':')[1]} -i \"{enc}\" -c copy \"{dest}\"";
        }

        var isSuccess = await RunCommandAsync(bin, cmd, workDir);
        
        // mp4decrypt 还原文件改名操作
        if (workDir is not null)
        {
            if (File.Exists(tmpEncFile)) File.Move(tmpEncFile, source);
            if (File.Exists(tmpDecFile)) File.Move(tmpDecFile, dest);
        }

        if (isSuccess)
        {
            if (tmpFile != "" && File.Exists(tmpFile)) File.Delete(tmpFile);
            return true;
        }
        
        Logger.Error(ResString.decryptionFailed);
        return false;
    }

    private static async Task<bool> RunCommandAsync(string name, string arg, string? workDir = null)
    {
        Logger.DebugMarkUp($"FileName: {name}");
        Logger.DebugMarkUp($"Arguments: {arg}");
        var process = Process.Start(new ProcessStartInfo()
        {
            FileName = name,
            Arguments = arg,
            // RedirectStandardOutput = true,
            // RedirectStandardError = true,
            CreateNoWindow = true,
            UseShellExecute = false,
            WorkingDirectory = workDir
        });
        await process!.WaitForExitAsync();
        return process.ExitCode == 0;
    }

    /// <summary>
    /// 从文本文件中查询KID的KEY
    /// </summary>
    /// <param name="file">文本文件</param>
    /// <param name="kid">目标KID</param>
    /// <returns></returns>
    public static async Task<string?> SearchKeyFromFileAsync(string? file, string? kid)
    {
        try
        {
            if (string.IsNullOrEmpty(file) || !File.Exists(file) || string.IsNullOrEmpty(kid)) 
                return null;

            Logger.InfoMarkUp(ResString.searchKey);
            using var stream = new FileStream(file, FileMode.Open, FileAccess.Read, FileShare.Read);
            using var reader = new StreamReader(stream);
            while (await reader.ReadLineAsync() is { } line)
            {
                if (!line.Trim().StartsWith(kid)) continue;
                
                Logger.InfoMarkUp($"[green]OK[/] [grey]{line.Trim()}[/]");
                return line.Trim();
            }
        }
        catch (Exception ex)
        {
            Logger.ErrorMarkUp(ex.Message);
        }
        return null;
    }

    public static ParsedMP4Info GetMP4Info(byte[] data)
    {
        var info = MP4InitUtil.ReadInit(data);
        if (info.Scheme != null) Logger.WarnMarkUp($"[grey]Type: {info.Scheme}[/]");
        if (info.PSSH != null) Logger.WarnMarkUp($"[grey]PSSH(WV): {info.PSSH}[/]");
        if (info.KID != null) Logger.WarnMarkUp($"[grey]KID: {info.KID}[/]");
        return info;
    }

    public static ParsedMP4Info GetMP4Info(string output)
    {
        using var fs = File.OpenRead(output);
        var header = new byte[1 * 1024 * 1024]; // 1MB
        _ = fs.Read(header);
        return GetMP4Info(header);
    }

    public static string? ReadInitShaka(string output, string bin)
    {
        Regex shakaKeyIdRegex = KidOutputRegex();

        // TODO: handle the case that shaka packager actually decrypted (key ID == ZeroKid)
        //       - stop process
        //       - remove {output}.tmp.webm
        var cmd = $"--quiet --enable_raw_key_decryption input=\"{output}\",stream=0,output=\"{output}.tmp.webm\" " +
                  $"--keys key_id={ZeroKid}:key={ZeroKid}";

        using var p = new Process();
        p.StartInfo = new ProcessStartInfo()
        {
            FileName = bin,
            Arguments = cmd,
            RedirectStandardOutput = true,
            RedirectStandardError = true,
            UseShellExecute = false
        };
        p.Start();
        var errorOutput = p.StandardError.ReadToEnd();
        p.WaitForExit();
        return shakaKeyIdRegex.Match(errorOutput).Groups[1].Value;
    }

    [GeneratedRegex("Key for key_id=([0-9a-f]+) was not found")]
    private static partial Regex KidOutputRegex();
}